/* reverb.c */
/* Placed in public domain by Steve Kirkendall.  Use at your own risk */

/* This file contains the reverb() function, which compute the reverb signal
 * for a given sound.
 */

#include <stdlib.h>
#include "unixsnd.h"
#define QTY(array)	(sizeof(array) / sizeof(*(array)))

/* Reverb uses a multistage digital filter with some long delays and some
 * short delays.  This table stores those delays, and the multiplier to use
 * then with.
 */
typedef struct
{
	int	weight;	/* multiplier for this delay */
	int	delay;	/* delay, in 100 microsecond increments */
	int	samples;/* delay, in samples (computed at runtime) */
} filter_t;

/* These are a variety of reverb filters */
static filter_t large[] = {
	{ 50, 1000},
	{-45, 1005},
	{ 45, 1200},
	{-40, 1210},
	{ 40, 1400},
	{-20, 1412},
	0
};
static filter_t small[] = {
	{ 30,   20},
	{-30,   23},
	{ 25,   34},
	{-15,   50},
	{ 60,   62},
	{-60,   68},
	{ 15,   79},
	{-15,   91},
	{ 40,  500},
	{ 20, 1000},
	{ 10, 1500},
	0
};
static filter_t echo[] = {
	{100, 5000},
	0
};
static filter_t multiecho[] = {
	{ 80, 1000},
	{ 60, 2000},
	{ 40, 3000},
	{ 20, 4000},
	0
};
static filter_t bandpass[] = {
	{ 60,   10},
	{-50,   15},
	{ 40,   20},
	{-30,   25},
	{ 20,   30},
	{-10,   35},
	0
};


/* pointer to current filter -- default is the "large" filter */
static filter_t *filter = large;



/* This array is used to assign names to the filters */
static struct
{
	char	*name;
	filter_t *filter;
} allfilters[] = {
	{"large", large},
	{"small", small},
	{"echo", echo},
	{"multiecho", multiecho},
	{"bandpass", bandpass},
};

/* These store old sound data which may still influence the reverb */
static mixed_t *oldsample;
static int nolds;
static int nextold;
static int usedold;	/* number of oldsample entries used for silence */


/* This stores the current reverb level, as a number from 0 to 100 */
static int rvbleft, rvbright;


/* Set up the reverb filter for the current output sample rate */
void reverbinit(char *value)
{
	int	i, j, n;
	int	extra;	/* number of back samples to store */
	char	name[30];
	filter_t *newfilter;
	mixed_t	*newsample;

	/* parse the reverb level */
	*name = '\0';
	if (sscanf(value, "%d,%d", &i, &j) != 2
	 && sscanf(value, "%[a-z]%d,%d", name, &i, &j) != 3)
	{
		if (sscanf(value, "%d", &i) != 1
		 && sscanf(value, "%[a-z]%d", name, &i) != 2)
		{
			if (verbose)
				printf("can't parse reverb value; ignoring\n");
		}
		j = i;
	}

	/* ignore bad values */
	newfilter = filter;
	if (*name)
	{
		for (n = 0; n < QTY(allfilters); n++)
			if (!strcmp(allfilters[n].name, name))
				break;
		if (n == QTY(allfilters))
		{
			if (verbose)
				printf("unknown reverb filter \"%s\"\n", name);
			return;
		}
		newfilter = allfilters[n].filter;
	}
	if (i < 0 || j < 0 || i > 5000 || j > 5000)
	{
		if (verbose)
			printf("bad reverb level\n");
		return;
	}

	/* if same reverb level, then do nothing */
	if (rvbleft == i && rvbright == j && newfilter == filter)
		return;
	rvbleft = i;
	rvbright = j;
	filter = newfilter;

	/* convert the filter's delay values to samples */
	for (j = extra = 0; filter[j].weight; j++)
	{
		filter[j].samples = filter[j].delay * output_frequency / 10000;
		if (filter[j].samples > extra)
			extra = filter[j].samples;
	}

	/* allocate the extra samples */
	if (nolds != extra)
	{
		newsample = (mixed_t *)malloc(extra * sizeof(mixed_t));
		memset(newsample, 0, extra * sizeof(mixed_t));
		if (oldsample)
		{
			for (i = 0, j = nolds - usedold; j > 0; j--)
			{
				newsample[i] = oldsample[nextold];
				if (++nextold == nolds)
					nextold = 0;
				if (++i == extra)
					i = 0;
			}
			free(oldsample);
			usedold += extra - nolds;
			if (usedold > 0)
				usedold = 0;
			nextold = i;
		}
		oldsample = newsample;
		nolds = extra;
	}
	else if (rvbleft == 0 && rvbright == 0)
	{
		memset(oldsample, 0, extra * sizeof(mixed_t));
	}
}

/* Return the reverb signal during a period of silence. */
int reverbsilence(mixed_t *sample, int maxsamples)
{
	int	i, j, n;
	long	left, right;

	/* if reverb is off or all reverb echoes used, then return nothing */
	if (rvbleft + rvbright == 0 || usedold == nolds)
		return 0;

	/* build the reverb data from previous sounds; assume there is no
	 * new sound yet.
	 */
	for (i = 0; usedold < nolds && i < maxsamples; i++)
	{
		/* compute the reverb signal */
		left = right = 0L;
		for (j = 0; filter[j].weight; j++)
		{
			n = nextold - filter[j].samples;
			if (n < 0)
				n += nolds;
			left += filter[j].weight * oldsample[n].left;
			right += filter[j].weight * oldsample[n].right;
		}

		/* zero the oldsample data at this sample position */
		oldsample[nextold].left = oldsample[nextold].right = 0;
		if (++nextold == nolds)
			nextold = 0;
		usedold++;

		/* store the reverb signal */
		sample[i].left = left * rvbleft / 100;
		sample[i].right = right * rvbright / 100;
	}

	/* return the number of samples */
	return i;
}

/* Mix reverb into a signal, and also store a copy of the signal for later
 * reverb computations.
 */
void reverbsamples(mixed_t *sample, int nsamples)
{
	int	i, j, n;
	int	extra;
	long	left, right;

	/* if reverb is off, then do nothing */
	if (rvbleft + rvbright == 0 || nsamples == 0)
		return;

	/* Apply the reverb filter to the new data.  Also copy the new samples
	 * into the oldsample buffer, for use in later sounds. */
	for (i = 0; i < nsamples; i++)
	{
		/* compute the filtered signal */
		left = right = 0L;
		for (j = 0; filter[j].weight; j++)
		{
			n = nextold - filter[j].samples;
			if (n < 0)
				n += nolds;
			left += filter[j].weight * oldsample[n].left;
			right += filter[j].weight * oldsample[n].right;
		}

		/* save the raw signal, to be used in later filtering */
		oldsample[nextold].left = sample[i].left / 100;
		oldsample[nextold].right = sample[i].right / 100;
		if (++nextold == nolds)
			nextold = 0;

		/* add the filtered signal to the raw signal */
		sample[i].left += left * rvbleft / 100;
		sample[i].right += right * rvbright / 100;
	}
	usedold = 0;
}


/* Return 1 if reverb is idle (has no echoes to fill silence with) */
int reverbidle(void)
{
	return usedold == nolds;
}
